@@ -1,20 +1,41 @@
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 from inspect import getfullargspec
 import torch
 from torch.autograd import Function, grad
-from torchcde import NaturalCubicSpline, natural_cubic_coeffs
-from torchdyn.numerics.odeint import odeint
-import torchdiffeq
+from torchcde import CubicSpline, natural_cubic_coeffs
+from torchdyn.numerics.odeint import odeint, odeint_mshooting
+
+
+def generic_odeint(problem_type, vf, x, t_span, solver, atol, rtol, interpolator, B0=None, 
+                  return_all_eval=False, maxiter=4, fine_steps=4):
+    "Dispatches to appropriate `odeint` function depending on `Problem` class (ODEProblem, MultipleShootingProblem)"
+    if problem_type == 'standard':
+        return odeint(vf, x, t_span, solver, atol=atol, rtol=rtol, interpolator=interpolator, return_all_eval=return_all_eval)
+    elif problem_type == 'multiple_shooting':
+        return odeint_mshooting(vf, x, t_span, solver, B0=B0, fine_steps=fine_steps, maxiter=maxiter)
 
 
 # TODO: optimize and make conditional gradient computations w.r.t end times
 # TODO: link `seminorm` arg from `ODEProblem`
-def _gather_odefunc_adjoint(vf, vf_params, solver, atol, rtol, interpolator,
-                            solver_adjoint, atol_adjoint, rtol_adjoint, integral_loss):
+def _gather_odefunc_adjoint(vf, vf_params, solver, atol, rtol, interpolator, solver_adjoint, 
+                            atol_adjoint, rtol_adjoint, integral_loss, problem_type, maxiter=4, fine_steps=4):
     "Prepares definition of autograd.Function for adjoint sensitivity analysis of the above `ODEProblem`"
     class _ODEProblemFunc(Function):
         @staticmethod
-        def forward(ctx, vf_params, x, t_span):
-            t_sol, sol = odeint(vf, x, t_span, solver, atol=atol, rtol=rtol, interpolator=interpolator)
+        def forward(ctx, vf_params, x, t_span, B=None):
+            t_sol, sol = generic_odeint(problem_type, vf, x, t_span, solver, atol, rtol, interpolator, B, 
+                                        False, maxiter, fine_steps)
             ctx.save_for_backward(sol, t_sol)
             return t_sol, sol
 
@@ -61,7 +82,6 @@ def _gather_odefunc_adjoint(vf, vf_params, solver, atol, rtol, interpolator,
                                       solver_adjoint, atol=atol_adjoint, rtol=rtol_adjoint,
                                       seminorm=(True, xT_nel+λT_nel))
                 # prepare adjoint state for next interval
-
                 #TODO: reuse vf_eval for dLdt calculations
                 xt = A[-1, :xT_nel].reshape(xT_shape)
                 dLdt_ = A[-1, xT_nel:xT_nel + λT_nel]@vf(t_sol[i], xt).flatten()
@@ -74,22 +94,20 @@ def _gather_odefunc_adjoint(vf, vf_params, solver, atol, rtol, interpolator,
             λ, μ = A[xT_nel:xT_nel + λT_nel], A[-μT_nel-1:-1]
             λ, μ = λ.reshape(λT.shape), μ.reshape(μT.shape)
             λ_tspan = torch.stack([dLdt[0], dLdt[-1]])
-            return (μ, λ, λ_tspan, None, None)
+            return (μ, λ, λ_tspan, None, None, None)
 
     return _ODEProblemFunc
 
 
 #TODO: introduce `t_span` grad as above
-#TODO: introduce option to interpolate on all solution points evaluated
-# TODO (qol): if `t_sol` contains additional solution points other than the specified ones in `t_span`
-# use those to interpolate
-def _gather_odefunc_interp_adjoint(vf, vf_params, solver, atol, rtol, interpolator,
-                                solver_adjoint, atol_adjoint, rtol_adjoint, integral_loss):
+def _gather_odefunc_interp_adjoint(vf, vf_params, solver, atol, rtol, interpolator, solver_adjoint, 
+                                   atol_adjoint, rtol_adjoint, integral_loss, problem_type, maxiter=4, fine_steps=4):
     "Prepares definition of autograd.Function for interpolated adjoint sensitivity analysis of the above `ODEProblem`"
     class _ODEProblemFunc(Function):
         @staticmethod
-        def forward(ctx, vf_params, x, t_span):
-            t_sol, sol = odeint(vf, x, t_span, solver, atol=atol, rtol=rtol, interpolator=interpolator, return_all_eval=True)
+        def forward(ctx, vf_params, x, t_span, B=None):
+            t_sol, sol = generic_odeint(problem_type, vf, x, t_span, solver, atol, rtol, interpolator, B, 
+                                        True, maxiter, fine_steps)
             ctx.save_for_backward(sol, t_span, t_sol)
             return t_sol, sol
 
@@ -103,8 +121,9 @@ def _gather_odefunc_interp_adjoint(vf, vf_params, solver, atol, rtol, interpolat
             λT_nel, μT_nel = λT.numel(), μT.numel()
             xT_shape, λT_shape, μT_shape = xT.shape, λT.shape, μT.shape
             A = torch.cat([λT.flatten(), μT.flatten()])
+
             spline_coeffs = natural_cubic_coeffs(x=sol.permute(1, 0, 2).detach(), t=t_sol)
-            x_spline = NaturalCubicSpline(coeffs=spline_coeffs, t=t_sol)
+            x_spline = CubicSpline(coeffs=spline_coeffs, t=t_sol)
 
             # define adjoint dynamics
             def adjoint_dynamics(t, A):
@@ -136,6 +155,6 @@ def _gather_odefunc_interp_adjoint(vf, vf_params, solver, atol, rtol, interpolat
 
             λ, μ = A[:λT_nel], A[-μT_nel:]
             λ, μ = λ.reshape(λT.shape), μ.reshape(μT.shape)
-            return (μ, λ, None, None)
+            return (μ, λ, None, None, None)
 
     return _ODEProblemFunc
\ No newline at end of file
